#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE.python-nubia file in the root directory of this source tree.
#

from .internal import context, exceptions
from .internal.deprecation import deprecated
from .internal.io import eventbus
from .internal.io.session_logger import SessionLogger
from .internal.nubia import Nubia
from .internal.options import Options
from .internal.plugin_interface import CompletionDataSource, PluginInterface
from .internal.typing import argument, command
from .internal.ui import statusbar

name = "nubia"

__all__ = [
    "CompletionDataSource",
    "Nubia",
    "Options",
    "PluginInterface",
    "SessionLogger",
    "argument",
    "command",
    "context",
    "deprecated",
    "eventbus",
    "exceptions",
    "statusbar",
]

__version__ = "0.3.0"
